package com.emmetgray.wrpn;

// This is the Big Integer class... it is somewhat analogous
// to the java.math.BigInteger class, except that this version
// has arbitrary precision plus a few calculator-specific
// functions
public class BigInt
{
    private boolean pfvOverflow;
    private boolean pfvCarryBit;
    private boolean pfvLossOfPrecision;
    private ArithMode pfvArithMode;

    private int bitSize;
    private int lastByte;
    private byte[] n;

    // The Arithmetic mode for signed decimal values
    public enum ArithMode
    {
        Unsigned,
        OnesComp,
        TwosComp
    }

    /*
     * **************************** Constructors ******************************
     */

    // The default constructor
    public BigInt()
    {
        pfvOverflow = false;
        pfvCarryBit = false;
        pfvLossOfPrecision = false;
        pfvArithMode = ArithMode.TwosComp;
        bitSize = 32;
        lastByte = 3;
        n = new byte[lastByte + 1];
    }

    // Constructor given a required size and mode
    public BigInt(int size, ArithMode mode)
    {
        this();
        pfvArithMode = mode;

        SetSize(size);
    }

    // Constructor from a byte array
    public BigInt(byte[] byte_array, int size, ArithMode mode)
    {
        this();
        pfvArithMode = mode;

        Import(byte_array, size, false);
    }

    // Constructor from another BigInt
    public BigInt(BigInt bi, int size, ArithMode mode)
    {
        this();
        pfvArithMode = mode;

        Import(bi.n, size, false);
    }

    // Constructor from a double
    public BigInt(double dbl, int size, ArithMode mode)
    {
        this();
        pfvArithMode = mode;

        byte[] bytes;
        long temp;

        // We actually just convert the double to a long, and then
        // perform the normal import for a long.

        // Convert only if there is some useful value
        if (dbl >= Long.MAX_VALUE || dbl <= Long.MIN_VALUE)
        {
            // Humm... should this generate an exception?
            pfvLossOfPrecision = true;
            return;
        }

        // v7.0.0 - 6 Jun 21: Fix 1's complement error
        if (dbl < 0 && mode == ArithMode.OnesComp)
        {
            dbl--;
        }

        // Use default rounding rules!
        temp = (long) dbl;
        bytes = GetBytes(temp);

        // only 15 digit precision for a double
        Import(bytes, size, (dbl < 0));
    }

    // Constructor from an integer
    public BigInt(int i, int size, ArithMode mode)
    {
        this();
        pfvArithMode = mode;

        byte[] bytes;

        // v7.0.0 - 6 Jun 21: Fix 1's complement error
        if (i < 0 && mode == ArithMode.OnesComp)
        {
            i--;
        }

        bytes = GetBytes(i);
        Import(bytes, size, (i < 0));
    }

    // Constructor from a long
    public BigInt(long ln, int size, ArithMode mode)
    {
        this();
        pfvArithMode = mode;

        byte[] bytes;

        // v7.0.0 - 6 Jun 21: Fix 1's complement error
        if (ln < 0 && mode == ArithMode.OnesComp)
        {
            ln--;
        }

        bytes = GetBytes(ln);
        Import(bytes, size, (ln < 0));
    }

    // Constructor from a string
    public BigInt(String s, int size, ArithMode mode)
    {
        this();
        int i, bits, temp, n_byte, n_bit, finish;
        String c, number;
        BigInt z;

        if (s == null || s.equals(""))
        {
            return;
        }

        // v7.0.1 - 3 Nov 21: Let's simplify calculation of the size
        z = new BigInt(size + 8, mode);
        this.SetSize(size + 8);
        pfvArithMode = mode;
        pfvLossOfPrecision = false;

        if (s.length() > 2 && s.startsWith("&"))
        {
            number = s.substring(2).toLowerCase().trim();
            switch (s.toLowerCase().toCharArray()[1])
            {
                case 'o': // Octal notation
                    finish = 0;
                    // input validation from right to left
                    for (i = number.length() - 1; i >= 0; i--)
                    {
                        // Stop at first unknown character
                        c = number.substring(i, i + 1);
                        if (!"01234567".contains(c))
                        {
                            finish = finish + 1;
                            break;
                        }
                    }

                    // Process the string
                    bits = 0;
                    for (i = number.length() - 1; i >= finish; i--)
                    {
                        temp = Integer.parseInt(number.substring(i, i + 1));
                        if (temp != 0)
                        {
                            // Sneaky trick to prevent multiplication/shifts
                            z.Clear();
                            z.PutAt(temp, bits);
                            _Add(z);
                        }
                        // Number of bits to shift
                        bits = bits + 3;
                        // v7.0.3 - 28 Jun 23: Trimming input during shortening bit size
                        if (bits > size)
                        {
                            break;
                        }
                    }
                    break;
                case 'b': // Binary (not a really a standard notation!)
                    finish = 0;
                    // input validation from right to left
                    for (i = number.length() - 1; i >= 0; i--)
                    {
                        // Stop at first unknown character
                        c = number.substring(i, i + 1);
                        if (!"01".contains(c))
                        {
                            finish = finish + 1;
                            break;
                        }
                    }
                    // v7.0.3 - 28 Jun 23: Trimming input during shortening bit size
                    if (finish > size)
                    {
                        finish = size;
                    }
                    // Process the string
                    for (i = number.length() - 1; i >= finish; i--)
                    {
                        // Which byte and which bit?
                        n_byte = (number.length() - 1 - i) / 8;
                        n_bit = (number.length() - 1 - i) % 8;
                        if (number.charAt(i) == '1')
                        {
                            this.n[n_byte] = (byte) (this.n[n_byte] + Math.pow(2, n_bit));
                        }
                    }
                    break;
                case 'h': // Hexadecimal notation
                    finish = 0;
                    // input validation from right to left
                    for (i = number.length() - 1; i >= 0; i--)
                    {
                        // Stop at first unknown character
                        c = number.substring(i, i + 1);
                        if (!"0123456789abcdef".contains(c))
                        {
                            finish = finish + 1;
                            break;
                        }
                    }

                    // Process the string
                    bits = 0;
                    for (i = number.length() - 1; i >= finish; i--)
                    {
                        temp = Integer.parseInt(number.substring(i, i + 1), 16);
                        if (temp != 0)
                        {
                            // Sneaky trick to prevent multiplication/shifts
                            z.Clear();
                            z.PutAt(temp, bits);
                            _Add(z);
                        }
                        // Number of bits to shift
                        bits = bits + 4;
                        // v7.0.3 - 28 Jun 23: Trimming input during shortening bit size
                        if (bits > size)
                        {
                            break;
                        }
                    }
                    break;
            }
        }
        else
        {
            // If no leading "&", it must be a decimal notation
            finish = 0;
            // input validation from right to left. This might seem a bit counterintuitive
            // since "123.45" would yield "45" and not "123"
            for (i = s.length() - 1; i >= 0; i--)
            {
                // Stop at first unknown character (including a decimal or
                // negative sign)
                c = s.substring(i, i + 1);
                if (!"0123456789".contains(c))
                {
                    finish = i + 1;
                    break;
                }
            }

            BigInt mult = new BigInt(size +8, mode);
            BigInt ten = new BigInt(size +8, mode);
            BigInt one = BigInt.One(size +8);

            // This is normal stuff. You multiply each digit times the value
            // associated with that decimal place, then add 'em all up.
            mult.n[0] = 1;
            ten.n[0] = 10;
            for (i = s.length() - 1; i >= finish; i--)
            {
                temp = Integer.parseInt(s.substring(i, i + 1));
                if (temp != 0)
                {
                    z.Clear();
                    z.n[0] = (byte) temp;
                    z.Multiply(mult, bitSize);
                    _Add(z);
                }
                // Update the Decimal place
                mult.Multiply(ten, bitSize);
                // Did you overflow the mult?
                if (mult.getOverflow())
                {
                    pfvLossOfPrecision = true;
                    break;
                }
            }
            // If it's negative assume 2's complement
            if (s.startsWith("-"))
            {
                OnesComplement(bitSize);
                // v7.0.0 - 6 Jun 21: Fix 1's complement error
                if (mode == ArithMode.TwosComp)
                {
                    _Add(one);
                }
            }
        }

        /*
         * We need to know if the input has exceeded the capacity of
         * the word size.  If so, then use the LossOfPrecision flag
         * so that ProcessKey can remove the last digit entered.
         */
        z = this.Copy();
        SetSize(size);
        // v7.0.1 - 3 Nov 21: Fixed a numeric entry "sign extension" bug
        if (z.MaxBit() != this.MaxBit())
        {
            pfvLossOfPrecision = true;
        }

        // v7.0.0 - 6 Jun 21: Fix 1's complement error
        // manually extend the sign
        if (s.startsWith("-") && mode == ArithMode.Unsigned)
        {
            for (int ext=lastByte; ext>0; ext--)
            {
                if (n[ext] == 0)
                {
                    n[ext] = (byte) 255;
                }
                else
                {
                    break;
                }
            }
        }
    }

    /*
     * **************************** Properties ******************************
     */

    // The Overflow property (read only)
    public boolean getOverflow()
    {
        return pfvOverflow;
    }

    // The Carry Bit property (read only)
    public boolean getCarryBit()
    {
        return pfvCarryBit;
    }

    // The Loss of Precision property (read only)
    public boolean getLossOfPrecision()
    {
        return pfvLossOfPrecision;
    }

    // The Big Integer ArithMode getter/setter
    public ArithMode getBIArithMode()
    {
        return pfvArithMode;
    }

    public void setBIArithMode(ArithMode mode)
    {
        pfvArithMode = mode;
    }

    // The Word Size getter/setter
    public int getWordSize()
    {
        return bitSize;
    }

    public void setWordSize(int size)
    {
        SetSize(size);
        bitSize = size;
    }

    // The Byte Array getter/setter
    public byte[] getByteArray()
    {
        return this.n;
    }

    /*
     * ************************** Math Operators ******************************
     * Note: No overloaded operators in Java
     */

    // Add two BigInt values
    public static BigInt Add(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.Add(y, 0);
        return ans;
    }

    // Add two BigInt values, with required size
    public static BigInt Add(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.Add(y, size);
        return ans;
    }

    // Add a BigInt value to the current value
    public void Add(BigInt y, int size)
    {
        int real_bitsize;
        BigInt temp_y;
        boolean sign_x, sign_y;

        temp_y = y.Copy();

        // determine the bitsize, and make 'em all the same
        real_bitsize = (size > 0) ? size : (Math.max(bitSize, temp_y.bitSize));
        SetSize(real_bitsize);
        temp_y.SetSize(real_bitsize);

        sign_x = this.SignBit();
        sign_y = temp_y.SignBit();

        // We temporarily need a bit more "wiggle room" than just the current
        // bitsize, so we add another byte. This is similar to SetSize, but
        // never does a sign extension
        AddPadding(RoundUp(real_bitsize));
        temp_y.AddPadding(bitSize);

        _Add(temp_y);

        // We needed the wiggle room to detect carry and overflow
        pfvCarryBit = MaxBit() >= real_bitsize;
        switch (getBIArithMode())
        {
            case Unsigned:
                // that was easy...
                pfvOverflow = pfvCarryBit;
                break;
            case OnesComp:
                // the "wrap around" rule for adding 1's complement numbers
                if (pfvCarryBit)
                {
                    _Add(BigInt.One(bitSize));
                }

                // determining overflow is a bit bizarre
                pfvOverflow = false;
                if (sign_x)
                {
                    if (sign_y && !(this.BitTest(real_bitsize - 1)) && pfvCarryBit)
                    {
                        pfvOverflow = true;
                    }
                }
                else if (!sign_y && this.BitTest(real_bitsize - 1) && !pfvCarryBit)
                {
                    pfvOverflow = true;
                }
                break;
            case TwosComp:
                // determining overflow is a bit bizarre
                pfvOverflow = false;
                if (sign_x)
                {
                    if (sign_y && !(this.BitTest(real_bitsize - 1)) && pfvCarryBit)
                    {
                        pfvOverflow = true;
                    }
                }
                else if (!sign_y && this.BitTest(real_bitsize - 1) && !pfvCarryBit)
                {
                    pfvOverflow = true;
                }
                break;
        }
        // OK, but it back to the real bitsize
        SetSize(real_bitsize);
    }

    // Internal version of Add (does not set Carry, Overflow, LOP)
    private void _Add(BigInt y)
    {
        int carry, temp;

        // Normal stuff... add two bytes, keep the "carry" for the next byte
        carry = 0;
        for (int i = 0; i <= this.lastByte; i++)
        {
            temp = (int) (this.n[i] & 0xff) + (int) (y.n[i] & 0xff) + carry;
            this.n[i] = (byte) (temp % 256);
            carry = temp / 256;
        }
        Mask();
    }

    // Subtract two BigInt values
    public static BigInt Subtract(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.Subtract(y, 0);
        return ans;
    }

    // Subtract two BigInt values with required size
    public static BigInt Subtract(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.Subtract(y, size);
        return ans;
    }

    // Subtract a BigInt value from the current value
    public void Subtract(BigInt y, int size)
    {
        int real_bitsize;
        BigInt temp_y;
        boolean sign_x, sign_y;

        temp_y = y.Copy();

        // determine the bitsize, and make 'em all the same
        real_bitsize = (size > 0) ? size : (Math.max(bitSize, temp_y.bitSize));
        SetSize(real_bitsize);
        temp_y.SetSize(real_bitsize);

        sign_x = this.SignBit();
        sign_y = temp_y.SignBit();

        // We temporarily need a bit more "wiggle room" than just the current
        // bitsize, so we add another byte. This is similar to SetSize, but
        // never does a sign extension
        AddPadding(RoundUp(real_bitsize));
        temp_y.AddPadding(bitSize);

        _Subtract(temp_y);

        // We needed the wiggle room to detect carry and overflow
        pfvCarryBit = (MaxBit() >= real_bitsize);
        switch (getBIArithMode())
        {
            case Unsigned:
                // that was easy...
                pfvOverflow = pfvCarryBit;
                break;
            case OnesComp:
                // the "wrap around" rule for subtracting 1's complement numbers
                if (pfvCarryBit)
                {
                    _Subtract(BigInt.One(bitSize));
                }
                // determining overflow is a bit bizarre
                pfvOverflow = false;
                if (sign_x)
                {
                    if (!sign_y && (!this.BitTest(real_bitsize - 1)) && !pfvCarryBit)
                    {
                        pfvOverflow = true;
                    }
                }
                else if (sign_y && this.BitTest(real_bitsize - 1) && pfvCarryBit)
                {
                    pfvOverflow = true;
                }
                break;
            case TwosComp:
                // determining overflow is a bit bizarre
                pfvOverflow = false;
                if (sign_x)
                {
                    if (!sign_y && (!this.BitTest(real_bitsize - 1)) && !pfvCarryBit)
                    {
                        pfvOverflow = true;
                    }
                }
                else if (sign_y && this.BitTest(real_bitsize - 1) && pfvCarryBit)
                {
                    pfvOverflow = true;
                }
                break;
        }
        // OK, but it back to the real bitsize
        SetSize(real_bitsize);
    }

    // Internal version of Subtract (does not set Carry, Overflow, LOP)
    private void _Subtract(BigInt y)
    {
        int carry, temp;

        // Normal stuff... subtract two bytes, generate a "borrow" if needed
        carry = 0;
        for (int i = 0; i <= this.lastByte; i++)
        {
            temp = (int) (this.n[i] & 0xff) - (int) (y.n[i] & 0xff) - carry;
            // Do we need a "borrow"?
            if (temp < 0)
            {
                temp = temp + 256;
                carry = 1;
            }
            else
            {
                carry = 0;
            }
            this.n[i] = (byte) temp;
        }
        Mask();
    }

    // Multiply two BigInt values
    public static BigInt Multiply(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.Multiply(y, 0);
        return ans;
    }

    // Multiply two BigInt values with required size
    public static BigInt Multiply(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.Multiply(y, size);
        return ans;
    }

    // Multiply a BigInt value with the current value
    public void Multiply(BigInt y, int size)
    {
        int temp, real_bitsize;
        BigInt temp_x, temp_y, z;
        boolean neg = false;

        temp_x = this.Copy();
        temp_y = y.Copy();

        // We need a bit more "wiggle room" than just the current bitsize,
        // so round off to the nearest byte and add 1 byte
        real_bitsize = (size > 0) ? size : (Math.max(bitSize, temp_y.bitSize));
        SetSize(RoundUp(real_bitsize));
        temp_x.SetSize(bitSize);
        temp_y.SetSize(bitSize);
        z = new BigInt(bitSize, pfvArithMode);
        Clear();

        // Let's handle some special cases
        if (temp_x.IsZero() || temp_y.IsZero())
        {
            SetSize(real_bitsize);
            return;
        }

        // v7.0.0 - 6 Jun 21: Fix 1's complement error
        if (pfvArithMode != ArithMode.Unsigned)
        {
            if (temp_x.SignBit() ^ temp_y.SignBit())
            {
                neg = true;
            }
            temp_x.AbsoluteValue();
            temp_y.AbsoluteValue();
        }

        // This is just like you were taught is school... multiply each digit
        // in the first number with every digit in the second. Keep track of
        // what "column" to write down the intermediate values, then add the
        // intermediate values together.
        for (int i = 0; i <= temp_x.MaxByte(); i++)
        {
            for (int j = 0; j <= temp_y.MaxByte(); j++)
            {
                temp = (int) (temp_x.n[i] & 0xff) * (int) (temp_y.n[j] & 0xff);
                if (temp != 0)
                {
                    if (i + j <= lastByte)
                    {
                        // This is a sneaky trick to create intermediate values
                        // without using shifts. Normally you'd multiply the
                        // temp value times the value for that "decimal place".
                        z.Clear();
                        z.PutAt(temp, (i + j) * 8);
                        _Add(z);
                    }
                    else
                    {
                        pfvOverflow = true;
                    }
                }
            }
        }
        // We needed the wiggle room to detect over_flow
        if (!pfvOverflow)
        {
            pfvOverflow = (MaxBit() >= real_bitsize);
        }
        SetSize(real_bitsize);

        // v7.0.0 - 6 Jun 21: Fix 1's complement error
        if (neg)
        {
            ChangeSign();
        }
    }

    // Divide two BigInt values
    public static BigInt Divide(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.Divide(y, 0);
        return ans;
    }

    // Divide two BigInt values with required size
    public static BigInt Divide(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.Divide(y, size);
        return ans;
    }

    // Divide the current value with a BigInt value
    public void Divide(BigInt y, int size)
    {
        int nx, ny, nc, ans_digit, real_bitsize;
        BigInt chunk, guess, temp_x, temp_y, z;
        boolean neg = false;

        temp_y = y.Copy();
        temp_x = this.Copy();

        // We need a bit more "wiggle room" than just the current bitsize,
        // so round off to the nearest byte and add 1 byte
        real_bitsize = (size > 0) ? size : (Math.max(temp_x.bitSize, temp_y.bitSize));
        SetSize(RoundUp(real_bitsize));
        temp_x.SetSize(bitSize);
        temp_y.SetSize(bitSize);
        guess = new BigInt(bitSize, pfvArithMode);
        chunk = new BigInt(bitSize, pfvArithMode);

        // Sanity check
        if (temp_y.IsZero())
        {
            pfvOverflow = true;
            SetSize(real_bitsize);
            Clear();
            return;
        }
        // When using signed numbers, we first figure out the what the sign of
        // the answer will be, then convert everything to absolute value
        if (pfvArithMode != ArithMode.Unsigned)
        {
            // if either is negative (but not both)
            if (temp_x.SignBit() ^ temp_y.SignBit())
            {
                neg = true;
            }
            temp_x.AbsoluteValue();
            temp_y.AbsoluteValue();
        }

        // Special case
        if (temp_x.IsLess(temp_y))
        {
            if (!temp_x.IsZero())
            {
                pfvCarryBit = true;
            }
            SetSize(real_bitsize);
            Clear();
            return;
        }

        // This is almost like you were taught in school... For each digit in
        // the answer, guess the number of times the divisor will go into a
        // "chunk" of the dividend. We know the size of the chunk will have
        // the same number of digits as x or be one digit larger. Subtract
        // the intermediate value and repeat.
        Clear();
        ny = temp_y.MaxByte();
        do
        {
            if (temp_x.IsLess(temp_y))
            {
                break;
            }

            nx = temp_x.MaxByte();
            // Build a "chunk" that has the same number of digits as y
            chunk.Clear();
            for (int i = 0; i <= ny; i++)
            {
                chunk.n[i] = temp_x.n[i + nx - ny];
            }
            nc = ny;

            // If it is too small, build another chunk with one more digit
            if (chunk.IsLess(temp_y))
            {
                for (int i = 0; i <= (ny + 1); i++)
                {
                    chunk.n[i] = temp_x.n[i + nx - ny - 1];
                }
                nc = (ny + 1);
            }

            // To keep from testing all 255 possibilities for each "answer
            // digit", we use a simplified binary tree search algorithm.
            ans_digit = 128;
            guess.Clear();
            for (int i = 6; i >= 0; i--)
            {
                guess.n[0] = (byte) ans_digit;
                z = Multiply(temp_y, guess);
                if (z.IsGreater(chunk))
                {
                    ans_digit = ans_digit - (int) Math.pow(2, i);
                }
                else
                {
                    ans_digit = ans_digit + (int) Math.pow(2, i);
                }
            }
            guess.n[0] = (byte) ans_digit;
            z = Multiply(temp_y, guess);
            if (z.IsGreater(chunk))
            {
                ans_digit = ans_digit - 1;
            }

            // Place answer digit in the proper "column"
            z.Clear();
            z.PutAt(ans_digit, (nx - nc) * 8);

            // Add it to the answer.
            _Add(z);

            // Generate the intermediate value by shifting the "last guess"
            // multiplier to its proper place (which undoes the effect of the
            // shift that occurred when we developed the chunk).
            guess.Clear();
            guess.n[nx - nc] = (byte) ans_digit;
            z = Multiply(temp_y, guess);

            // Now subtract the intermediate value and continue...
            temp_x._Subtract(z);
        } while (nx - nc > 0);

        pfvCarryBit = !temp_x.IsZero();
        pfvOverflow = (MaxBit() >= real_bitsize);
        SetSize(real_bitsize);

        if (neg)
        {
            ChangeSign();
        }
    }

    // Modulus (Division Remainder)
    public static BigInt Mod(BigInt x, BigInt y)
    {
        return Remainder(x, y, 0);
    }

    // Modulus (Division Remainder) with required size
    public static BigInt Mod(BigInt x, BigInt y, int size)
    {
        return Remainder(x, y, size);
    }

    // Modulus (Division Remainder)
    public static BigInt Remainder(BigInt x, BigInt y)
    {
        return Remainder(x, y, 0);
    }

    // Modulus (Division Remainder) with required size
    public static BigInt Remainder(BigInt x, BigInt y, int size)
    {
        int nx, ny, nc, ans_digit, real_bitsize;
        BigInt chunk, guess, temp_x, temp_y, z, ans;
        boolean neg = false;

        temp_x = x.Copy();
        temp_y = y.Copy();

        // We need a bit more wiggle room than the just current bitsize.
        real_bitsize = size > 0 ? size : Math.max(temp_x.bitSize, temp_y.bitSize);
        ans = new BigInt(RoundUp(real_bitsize), x.pfvArithMode);
        temp_x.SetSize(ans.bitSize);
        temp_y.SetSize(ans.bitSize);
        guess = new BigInt(ans.bitSize, x.pfvArithMode);
        chunk = new BigInt(ans.bitSize, x.pfvArithMode);

        // Sanity check
        if (temp_y.IsZero())
        {
            ans.SetSize(real_bitsize);
            ans.Clear();
            ans.pfvOverflow = true;
            return ans;
        }

        // When using signed numbers, we first figure out the what the sign of
        // the answer will be, then convert everything to absolute value
        if (x.pfvArithMode != ArithMode.Unsigned)
        {
            if (x.SignBit())
            {
                neg = true;
            }
            temp_x.AbsoluteValue();
            temp_y.AbsoluteValue();
        }

        // Special case (if x < y, then return x)
        if (temp_x.IsLess(temp_y))
        {
            temp_x.SetSize(real_bitsize);
            if (neg)
            {
                temp_x.ChangeSign();
            }
            return temp_x;
        }

        // This is almost like you were taught in school... For each digit in
        // the answer, guess the number of times the divisor will go into a
        // "chunk" of the dividend. We know the size of the chunk will have
        // the same number of digits as x or be one digit larger. Subtract
        // the intermediate value and repeat.

        ans = temp_x.Copy();
        ny = temp_y.MaxByte();
        do
        {
            // If no more divisors... we're done
            if (ans.IsLess(temp_y))
            {
                break;
            }

            nx = ans.MaxByte();
            // Build a "chunk" that has the same number of digits as y
            chunk.Clear();
            for (int i = 0; i <= ny; i++)
            {
                chunk.n[i] = ans.n[i + nx - ny];
            }
            nc = ny;

            // If it is too small, build another chunk with one more digit
            if (chunk.IsLess(temp_y))
            {
                for (int i = 0; i <= ny + 1; i++)
                {
                    chunk.n[i] = ans.n[i + nx - ny - 1];
                }
                nc = ny + 1;
            }

            // To keep from testing all 255 possibilities for each "answer
            // digit", we use a simplified binary tree search algorithm.
            ans_digit = 128;
            guess.Clear();
            for (int i = 6; i >= 0; i--)
            {
                guess.n[0] = (byte) ans_digit;
                // This is why we need more wiggle room... so we don't have
                // to deal with overflows during the guessing of the "answer
                // digit"
                z = Multiply(temp_y, guess);
                if (z.IsGreater(chunk))
                {
                    ans_digit = (int) (ans_digit - (Math.pow(2, i)));
                }
                else
                {
                    ans_digit = (int) (ans_digit + (Math.pow(2, i)));
                }
            }
            guess.n[0] = (byte) ans_digit;
            z = Multiply(temp_y, guess);
            if (z.IsGreater(chunk))
            {
                ans_digit = ans_digit - 1;
            }

            // Generate the intermediate value by shifting the "last guess"
            // multiplier to its proper place (which undoes the effect of the
            // shift that occurred when we developed the chunk).
            guess.Clear();
            guess.n[nx - nc] = (byte) ans_digit;
            z = Multiply(temp_y, guess);

            // Now subtract the intermediate value and continue...
            ans._Subtract(z);
        } while (nx - nc > 0);

        ans.SetSize(real_bitsize);
        if (neg)
        {
            ans.ChangeSign();
        }
        return ans;
    }

    // Power (raise x to the power of y)
    public static BigInt Power(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.Power(y, 0);
        return ans;
    }

    // Power (raise x to the power of y) with required size
    public static BigInt Power(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.Power(y, size);
        return ans;
    }

    // Power (raise the current value to the power of y)
    public void Power(BigInt y, int size)
    {
        int real_bitsize;
        boolean over = false;
        BigInt temp_x, temp_y, one;

        temp_x = this.Copy();
        temp_y = y.Copy();

        // set to the correct size first, then add padding
        real_bitsize = (size > 0) ? size : (Math.max(temp_x.bitSize, temp_y.bitSize));
        SetSize(RoundUp(real_bitsize));

        temp_x.SetSize(bitSize);
        temp_y.SetSize(bitSize);
        one = BigInt.One(bitSize);

        // There are a whole bunch of special cases!
        if (temp_y.IsZero())
        {
            // Any number raised to the 0th power is 1
            Clear();
            n[0] = 1;
            SetSize(real_bitsize);
            return;
        }
        if (temp_y.Equals(one))
        {
            // any number raised to 1 is itself
            SetSize(real_bitsize);
            return;
        }
        if (temp_x.Equals(one))
        {
            // any power of 1 is 1
            SetSize(real_bitsize);
            return;
        }
        if (IsZero() || (temp_y.SignBit() && (pfvArithMode != ArithMode.Unsigned)))
        {
            // zero or negative
            SetSize(real_bitsize);
            Clear();
            return;
        }

        // Just multiply x with itself that many times
        temp_y._Subtract(one);
        while (!temp_y.IsZero())
        {
            Multiply(temp_x, bitSize);
            if (pfvOverflow)
            {
                over = true;
            }
            temp_y._Subtract(one);
        }
        pfvOverflow = (over | (MaxBit() >= real_bitsize));
        SetSize(real_bitsize);
    }

    // Square Root of x (using Newton's method)
    public static BigInt SquareRoot(BigInt x)
    {
        BigInt ans = x.Copy();
        ans.SquareRoot(0);
        return ans;
    }

    // Square Root of x (using Newton's method) with required size
    public static BigInt SquareRoot(BigInt x, int size)
    {
        BigInt ans = x.Copy();
        ans.SquareRoot(size);
        return ans;
    }

    // Square Root of the current value (using Newton's method)
    public void SquareRoot(int size)
    {
        BigInt temp_x, r;
        int real_bitsize;
        boolean sign;

        sign = SignBit();
        real_bitsize = size > 0 ? size : bitSize;
        SetSize(RoundUp(real_bitsize));

        // a quick sanity check
        if (IsZero() || (pfvArithMode != ArithMode.Unsigned) && sign)
        {
            pfvOverflow = true;
            Clear();
            return;
        }

        temp_x = this.Copy();
        r = new BigInt(bitSize, pfvArithMode);
        r.n[0] = 1;
        while (IsGreater(r))
        {
            r = Divide(temp_x, this);
            // Using right shift instead of division by 2
            _Add(r);
            RightShift(1, false, bitSize);
        }

        // Carry on remainder (not a perfect square)
        r = Multiply(this, this);
        r._Subtract(temp_x);
        pfvCarryBit = !r.IsZero();

        // remove the wiggle room (if any)
        SetSize(real_bitsize);
    }

    /*
     * ************************ Bitwise Operators ****************************
     */

    // Perform a Bitwise AND operation of x and y
    public static BigInt BitwiseAnd(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.BitwiseAnd(y, 0);
        return ans;
    }

    // Perform a Bitwise AND operation of x and y with required size
    public static BigInt BitwiseAnd(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.BitwiseAnd(y, size);
        return ans;
    }

    // Perform a Bitwise AND operation
    public void BitwiseAnd(BigInt y, int size)
    {
        BigInt temp_y;

        // First we have to make a copy of the variables
        temp_y = y.Copy();

        // pick the bit size of the answer
        SetSize((size > 0) ? size : (Math.max(bitSize, temp_y.bitSize)));

        // Resize the y (since the size may have changed)
        temp_y.SetSize(bitSize);

        for (int i = 0; i <= lastByte; i++)
        {
            this.n[i] = (byte) (this.n[i] & temp_y.n[i]);
        }
    }

    // Perform a Bitwise OR operation of x and y
    public static BigInt BitwiseOr(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.BitwiseOr(y, 0);
        return ans;
    }

    // Perform a Bitwise OR operation of x and y with required size
    public static BigInt BitwiseOr(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.BitwiseOr(y, size);
        return ans;
    }

    // Perform a Bitwise OR operation
    public void BitwiseOr(BigInt y, int size)
    {
        BigInt temp_y;

        temp_y = y.Copy();

        SetSize((size > 0) ? size : (Math.max(bitSize, temp_y.bitSize)));
        temp_y.SetSize(bitSize);

        for (int i = 0; i <= lastByte; i++)
        {
            this.n[i] = (byte) (this.n[i] | temp_y.n[i]);
        }
    }

    // Perform a Bitwise NOT operation of x
    public static BigInt OnesComplement(BigInt x)
    {
        BigInt ans = x.Copy();
        ans.OnesComplement(0);
        return ans;
    }

    // Perform a Bitwise NOT operation of x with required size
    public static BigInt OnesComplement(BigInt x, int size)
    {
        BigInt ans = x.Copy();
        ans.OnesComplement(size);
        return ans;
    }

    // Perform a Bitwise NOT operation
    public void OnesComplement(int size)
    {
        if (size > 0)
        {
            SetSize(size);
        }

        for (int i = 0; i <= lastByte; i++)
        {
            this.n[i] = (byte) (255 - this.n[i]);
        }
        Mask();
    }

    // Perform a Bitwise Exclusive OR operation on x and y
    public static BigInt Xor(BigInt x, BigInt y)
    {
        BigInt ans = x.Copy();
        ans.Xor(y, 0);
        return ans;
    }

    // Perform a Bitwise Exclusive OR operation on x and y with required size
    public static BigInt Xor(BigInt x, BigInt y, int size)
    {
        BigInt ans = x.Copy();
        ans.Xor(y, size);
        return ans;
    }

    // Perform a Bitwise Exclusive OR operation
    public void Xor(BigInt y, int size)
    {
        BigInt temp_y;

        temp_y = y.Copy();

        SetSize((size > 0) ? size : (Math.max(bitSize, temp_y.bitSize)));
        temp_y.SetSize(bitSize);

        for (int i = 0; (i <= lastByte); i++)
        {
            this.n[i] = (byte) (this.n[i] ^ temp_y.n[i]);
        }
    }

    // Shift the bit pattern of x to the Left
    public static BigInt LeftShift(BigInt x, int distance)
    {
        BigInt ans = x.Copy();
        ans.LeftShift(distance, 0);
        return ans;
    }

    // Shift the bit pattern of x to the Left with required size
    public static BigInt LeftShift(BigInt x, int distance, int size)
    {
        BigInt ans = x.Copy();
        ans.LeftShift(distance, size);
        return ans;
    }

    // Shift the bit pattern to the Left
    public void LeftShift(int distance, int size)
    {
        int temp, carry;
        BigInt temp_x;

        if (size > 0)
        {
            SetSize(size);
        }

        // a bit of sanity checking
        if (distance <= 0)
        {
            return;
        }

        // Shift left is the same as "multiply by 2"
        for (int j = 1; j <= distance; j++)
        {
            temp_x = this.Copy();

            // Carry is based upon the left most bit
            pfvCarryBit = SignBit();

            carry = 0;
            for (int i = 0; i <= lastByte; i++)
            {
                temp = ((int) (temp_x.n[i] & 0xff) * 2) + carry;
                this.n[i] = (byte) (temp % 256);
                carry = temp / 256;
            }
        }
        Mask();
    }

    // Shift the bit pattern of x to the Right
    public static BigInt RightShift(BigInt x, int distance)
    {
        BigInt ans = x.Copy();
        ans.RightShift(distance, false, 0);
        return ans;
    }

    // Shift the bit pattern of x to the Right with required size
    public static BigInt RightShift(BigInt x, int distance, int size)
    {
        BigInt ans = x.Copy();
        ans.RightShift(distance, false, size);
        return ans;
    }

    // Shift the bit pattern to the Right
    public void RightShift(int distance, boolean SaveSign, int size)
    {
        int carry;
        boolean sign_bit;
        BigInt temp_x;

        if (size > 0)
        {
            SetSize(size);
        }

        // a bit of sanity checking
        if (distance <= 0)
        {
            return;
        }

        for (int j = 1; j <= distance; j++)
        {
            temp_x = this.Copy();

            // Carry is based upon the right most bit
            pfvCarryBit = (temp_x.n[0] & 1) > 0;
            sign_bit = SignBit();

            // Shift right is the same as "divide by 2"
            carry = 0;
            for (int i = lastByte; i >= 0; i--)
            {
                if (i < lastByte)
                {
                    if ((temp_x.n[i + 1] & 1) != 0)
                    {
                        carry = 128;
                    }
                    else
                    {
                        carry = 0;
                    }
                }
                this.n[i] = (byte) (((int) (temp_x.n[i] & 0xff) / 2) | carry);
            }
            if (SaveSign & (sign_bit && (pfvArithMode != ArithMode.Unsigned)))
            {
                // are we doing an Arithmetic Shift Right?
                BitSet((bitSize - 1));
            }
        }
    }

    // Rotate the bit pattern of x to the Right
    public static BigInt RotateRight(BigInt x, int distance, boolean WithCarry,
                                     boolean StartingCarry, int size)
    {
        BigInt ans = x.Copy();
        ans.RotateRight(distance, WithCarry, StartingCarry, size);
        return ans;
    }

    // Rotate the bit pattern to the Right
    public void RotateRight(int distance, boolean WithCarry, boolean StartingCarry, int size)
    {
        int carry;
        BigInt temp_x;
        // v6.0.3 - 5 May 12

        if (size > 0)
        {
            SetSize(size);
        }

        // a bit of sanity checking
        if (distance <= 0)
        {
            return;
        }
        for (int j = 1; j <= distance; j++)
        {
            temp_x = this.Copy();

            // Carry is based upon the right most bit
            pfvCarryBit = (temp_x.n[0] & 1) != 0;

            carry = 0;
            for (int i = lastByte; i >= 0; i--)
            {
                if (i < lastByte)
                {
                    if ((temp_x.n[i + 1] & 1) != 0)
                    {
                        carry = 128;
                    }
                    else
                    {
                        carry = 0;
                    }
                }
                this.n[i] = (byte) (((int) (temp_x.n[i] & 0xff) / 2) | carry);
            }
            if (WithCarry)
            {
                if (StartingCarry)
                {
                    // add the sign bit back in
                    BitSet(bitSize - 1);
                }
            }
            else if ((temp_x.n[0] & 1) != 0)
            {
                // add the bit at position 0 back in
                BitSet(bitSize - 1);
            }
            StartingCarry = pfvCarryBit;
        }
    }

    // Rotate the bit pattern to the Left
    public static BigInt RotateLeft(BigInt x, int distance, boolean WithCarry, boolean StartingCarry, int size)
    {
        BigInt ans = x.Copy();
        ans.RotateLeft(distance, WithCarry, StartingCarry, size);
        return ans;
    }

    // Rotate the bit pattern to the Left
    public void RotateLeft(int distance, boolean WithCarry, boolean StartingCarry, int size)
    {
        int temp, carry;
        // v6.0.3 - 5 May 12

        if (size > 0)
        {
            SetSize(size);
        }

        // a bit of sanity checking
        if (distance <= 0)
        {
            return;
        }

        for (int j = 1; j <= distance; j++)
        {
            // Carry is based upon the left most bit
            pfvCarryBit = SignBit();

            // Rotate left is similar to "multiply by 2"
            carry = 0;
            for (int i = 0; i <= lastByte; i++)
            {
                temp = ((int) (this.n[i] & 0xff) * 2) + carry;
                this.n[i] = (byte) (temp % 256);
                carry = temp / 256;
            }
            if (WithCarry)
            {
                if (StartingCarry)
                {
                    // put the carry bit at position 0
                    this.BitSet(0);
                }
            }
            else if (pfvCarryBit)
            {
                // put the sign bit at position 0
                this.BitSet(0);
            }
            StartingCarry = pfvCarryBit;
        }
        Mask();
    }

    /*
     * ************************ Equality Operators ******************************
     */

    // Is the current BigInt equal to zero?
    public boolean IsZero()
    {
        for (int i = 0; i <= lastByte; i++)
        {
            if (this.n[i] != 0)
            {
                return false;
            }
        }
        // UNDONE: In some situations you might want 1's comp mode 0 = -0
        return true;
    }

    // Test for equality (is x = y)
    public static boolean Equals(BigInt x, BigInt y)
    {
        if (x == null & y == null)
        {
            return true;
        }

        if (x == null ^ y == null)
        {
            return false;
        }

        if (Compare(x, y) == 0)
        {
            return true;
        }
        return false;
    }

    // Test for equality
    public boolean Equals(BigInt y)
    {
        if (y == null)
        {
            return false;
        }

        if (Compare(this, y) == 0)
        {
            return true;
        }
        return false;
    }

    // Less Than
    public boolean IsLess(BigInt y)
    {
        if (Compare(this, y) == -1)
        {
            return true;
        }
        return false;
    }

    // Less Than or Equal
    public boolean IsLessOrEqual(BigInt y)
    {
        if (Compare(this, y) <= 0)
        {
            return true;
        }
        return false;
    }

    // Greater Than
    public boolean IsGreater(BigInt y)
    {
        if (Compare(this, y) == 1)
        {
            return true;
        }
        return false;
    }

    // Greater Than or Equal
    public boolean IsGreaterOrEqual(BigInt y)
    {
        if (Compare(this, y) >= 0)
        {
            return true;
        }
        return false;
    }

    // Compare a BigInt value with another
    public static int Compare(BigInt x, BigInt y)
    {
        BigInt temp_x, temp_y;

        // make copies since we may change the bit size
        temp_x = x.Copy();
        temp_y = y.Copy();

        if (temp_x.bitSize > temp_y.bitSize)
        {
            temp_y.SetSize(temp_x.bitSize);
        }
        if (temp_y.bitSize > temp_x.bitSize)
        {
            temp_x.SetSize(temp_y.bitSize);
        }

        // Comparing negative number is a bit easier, since we know that any
        // positive number is greater than any negative number
        if (x.pfvArithMode != ArithMode.Unsigned)
        {
            boolean sign_x = temp_x.SignBit();
            boolean sign_y = temp_y.SignBit();

            // when either is negative (but not both)
            if (sign_x ^ sign_y)
            {
                if (sign_y)
                {
                    return 1;
                }
                return -1;
            }
            // UNDONE: In some situations you might want 1's comp mode 0 = -0
        }
        for (int i = temp_x.lastByte; i >= 0; i--)
        {
            if ((int) (temp_x.n[i] & 0xff) > (int) (temp_y.n[i] & 0xff))
            {
                // x > y
                return 1;
            }
            if ((int) (temp_x.n[i] & 0xff) < (int) (temp_y.n[i] & 0xff))
            {
                // x < y
                return -1;
            }
        }
        // must be equal
        return 0;
    }

    /*
     * ************************* Import/Export functions *****************************
     */

    // Convert to an integer
    public int ToInteger()
    {
        BigInt x;

        x = this.Copy();
        x.SetSize(32);
        pfvLossOfPrecision = !Equals(x);

        // 2's complement mode rules apply!
        return GetInt(x.n);
    }

    // Convert to a Long
    public long ToLong()
    {
        BigInt x;

        x = this.Copy();
        x.SetSize(64);
        pfvLossOfPrecision = !Equals(x);

        // 2's complement mode rules apply!
        return GetLong(x.n);
    }

    // Convert to a double
    public double ToDouble()
    {
        // To Double (actually from long)
        return (double) this.ToLong();
    }

    // Override the default ToString() method
    public String ToString()
    {
        return ("&H" + ToStringHex());
    }

    // To Hexadecimal string
    public String ToStringHex()
    {
        int j, ending;
        StringBuilder sb = new StringBuilder();

        // How many digits are possible
        ending = (bitSize - 1) / 4;

        // Instead of using the normal method, we use the GetAt function to
        // extract the digit's value directly from the n() array.
        for (int i = 0; i <= ending; i++)
        {
            // Sneaky trick to prevent multiplication and division
            j = GetAt((i * 4), 4);
            if (j > 9)
            {
                // UNDONE: This will probably break globalization
                sb.insert(0, (char) ((int) ('a') + j - 10));
            }
            else
            {
                sb.insert(0, j);
            }
        }
        return sb.toString();
    }

    // To Binary string
    public String ToStringBin()
    {
        return ToStringBin(false);
    }

    // To Binary string
    public String ToStringBin(boolean pad)
    {
        StringBuilder sb = new StringBuilder();
        int spaces = 0;
        String ans;

        for (int i = lastByte; i >= 0; i--)
        {
            for (int j = 7; j >= 0; j--)
            {
                if (((byte) Math.pow(2, j) & n[i]) != 0)
                {
                    sb.append("1");
                }
                else
                {
                    sb.append("0");
                }
            }
            if ((i != 0) && (pad))
            {
                sb.append(" ");
                spaces++;
            }
        }

        // Trim to current word length, considering the optional spaces
        if (pad)
        {
            ans = sb.substring(sb.length() - (bitSize + spaces), sb.length());
        }
        else
        {
            ans = sb.substring(sb.length() - bitSize, sb.length());
        }
        return ans;
    }

    // To Octal String
    public String ToStringOct()
    {
        int j, ending;
        StringBuilder sb = new StringBuilder();

        // How many digits are possible
        ending = (bitSize - 1) / 3;

        // Instead of using the normal method, we use the GetAt function to
        // extract the digit's value directly from the n() array.
        for (int i = 0; i <= ending; i++)
        {
            // Sneaky trick to prevent multiplication and division
            j = GetAt((i * 3), 3);
            sb.insert(0, j);
        }
        return sb.toString();
    }

    // To Decimal String
    public String ToStringDec()
    {
        int j, ending;
        BigInt tempx, ten, z;
        boolean is_neg;
        StringBuilder sb = new StringBuilder();

        tempx = this.Copy();
        tempx.SetSize(RoundUp(bitSize));
        ten = new BigInt(tempx.bitSize, pfvArithMode);
        is_neg = false;

        switch (pfvArithMode)
        {
            case Unsigned:
                break;
            // Do nothing
            case OnesComp:
                if (SignBit())
                {
                    is_neg = true;
                    tempx.OnesComplement(tempx.bitSize);
                }
                break;
            case TwosComp:
                if (SignBit())
                {
                    is_neg = true;
                    tempx.OnesComplement(tempx.bitSize);
                    tempx._Add(BigInt.One(tempx.bitSize));
                }
                break;
        }

        // How many decimal digits are possible
        ending = (int) (Math.floor(this.bitSize / 3.33333333)) + 1;

        // This normal stuff... To get each digit, divide the number by 10
        // and then get the remainder. Repeat.
        ten.n[0] = 10;
        for (int i = 1; i <= ending; i++)
        {
            // Can't use my sneaky GetAt function... dang!
            z = Remainder(tempx, ten);
            j = z.n[0];
            sb.insert(0, j);
            tempx = Divide(tempx, ten);
        }

        if (is_neg)
        {
            sb.insert(0, "-");
        }

        return sb.toString();
    }

    /*
     * ************************ Misc Utilities *****************************
     */

    // The number 1 as a Big Integer
    public static BigInt One(int size)
    {
        BigInt ans = new BigInt(size, ArithMode.TwosComp);
        ans.n[0] = 1;
        return ans;
    }

    // Split a BigInt into two equal size words (even bit sizes only!)
    public BigInt[] Split() throws Exception
    {
        int half_bitsize, half_maxbyte;
        BigInt Right_Hand, Left_Hand;
        BigInt[] ans = new BigInt[2];

        // A wee bit of sanity checking
        if ((bitSize % 2) > 0)
        {
            throw new Exception("Bit size is not divisible by two");
        }

        half_bitsize = (int) (bitSize / 2);
        half_maxbyte = (half_bitsize - 1) / 8;
        Right_Hand = new BigInt(half_bitsize, pfvArithMode);
        Left_Hand = new BigInt(half_bitsize, pfvArithMode);

        // Right size (is easy)
        for (int i = 0; i <= half_maxbyte; i++)
        {
            Right_Hand.n[i] = this.n[i];
        }
        Right_Hand.Mask();

        // Left side (right shift "half_bitsize" number of times)
        BigInt temp_x = this.Copy();
        temp_x.RightShift(half_bitsize, false, bitSize);

        for (int i = 0; i <= half_maxbyte; i++)
        {
            Left_Hand.n[i] = temp_x.n[i];
        }
        Left_Hand.Mask();

        ans[0] = Right_Hand;
        ans[1] = Left_Hand;
        return ans;
    }

    // Create a bit mask
    public static BigInt CreateMask(int bits, boolean LeftJustify, int size, ArithMode mode)
    {
         BigInt ans = new BigInt(size, mode);
        if (LeftJustify)
        {
            for (int i = ans.bitSize - 1; i >= (ans.bitSize - bits); i--)
            {
                ans.BitSet(i);
            }
        }
        else
        {
            for (int i = 0; i <= (bits - 1); i++)
            {
                ans.BitSet(i);
            }
        }
        return ans;
    }

    // Move the bit pattern as many spaces as needed
    public static BigInt[] LeftJustify(BigInt x)
    {
        int d;
        BigInt Justified, distance;
        BigInt[] ans = new BigInt[2];

        Justified = x.Copy();
        d = x.bitSize - x.MaxBit() - 1;

        Justified.LeftShift(d, x.bitSize);
        distance = new BigInt(d, x.bitSize, x.pfvArithMode);

        ans[0] = Justified;
        ans[1] = distance;
        return ans;
    }

    // Convert x into its absolute value
    public static BigInt AbsoluteValue(BigInt x)
    {
        BigInt temp_x = x.Copy();
        temp_x.AbsoluteValue();
        return temp_x;
    }

    // Converts the number into its absolute value
    public void AbsoluteValue()
    {
        // Is there anything to do?
        if (!this.SignBit())
        {
            return;
        }
        switch (pfvArithMode)
        {
            case Unsigned:
                break;
            case OnesComp:
                OnesComplement(bitSize);
                break;
            case TwosComp:
                // Check for largest possible negative number in order
                // to set the Overflow flag
                BigInt sign = CreateMask(1, true, bitSize, ArithMode.Unsigned);
                if (Equals(sign))
                {
                    pfvOverflow = true;
                }
                else
                {
                    _Subtract(BigInt.One(bitSize));
                    OnesComplement(bitSize);
                }
                break;
        }
    }

    // Change the sign of the Big Integer
    public void ChangeSign()
    {
        if (SignBit())
        {
            // is already negative, so go to positive
            switch (pfvArithMode)
            {
                case Unsigned:
                    _Subtract(One(bitSize));
                    OnesComplement(bitSize);
                    // Not really an error, but we do set the overflow flag
                    pfvOverflow = true;
                    break;
                case OnesComp:
                    OnesComplement(bitSize);
                    break;
                case TwosComp:
                    _Subtract(One(bitSize));
                    OnesComplement(bitSize);
                    break;
            }
        }
        else
        {
            // is already positive, so go to negative
            switch (pfvArithMode)
            {
                case Unsigned:
                    OnesComplement(bitSize);
                    _Add(One(bitSize));
                    // Not really an error, but we do set the overflow flag
                    pfvOverflow = true;
                    break;
                case OnesComp:
                    OnesComplement(bitSize);
                    break;
                case TwosComp:
                    OnesComplement(bitSize);
                    _Add(One(bitSize));
                    break;
            }
        }
    }

    // Combine two halves into one value.
    public void Combine(BigInt Left_Hand, BigInt Right_Hand) throws Exception
    {
        BigInt temp;
        int i;

        // A wee bit of sanity checking
        if ((Left_Hand.bitSize != Right_Hand.bitSize))
        {
            throw new Exception("Bits sizes do not match");
        }

        SetSize((Left_Hand.bitSize * 2));
        Clear();

        // Right side
        for (i = 0; i <= Right_Hand.lastByte; i++)
        {
            n[i] = Right_Hand.n[i];
        }

        // Left side (shift left "left_hand.bitsize" number of times)
        temp = Left_Hand.Copy();
        temp.SetSize((Left_Hand.bitSize * 2));
        temp.LeftShift(Left_Hand.bitSize, 0);

        // combine with the existing
        for (i = 0; i <= temp.lastByte; i++)
        {
            n[i] = (byte) (n[i] | temp.n[i]);
        }
    }

    // Count the number of set bits.
    public void SumBits()
    {
        BigInt temp;
        BigInt one = BigInt.One(bitSize);

        temp = this.Copy();
        Clear();

        for (int i = lastByte; i >= 0; i--)
        {
            for (int j = 7; j >= 0; j--)
            {
                if ((temp.n[i] & (byte) Math.pow(2, j)) != 0)
                {
                    _Add(one);
                }
            }
        }
    }

    // Clear the bit at the given location
    public void BitClear(int location)
    {
        int n_byte;
        byte n_bit;
        n_byte = location / 8;

        n_bit = (byte) ~(int) (Math.pow(2, (location % 8)));

        if (n_byte <= lastByte)
        {
            this.n[n_byte] = (byte) (this.n[n_byte] & n_bit);
        }
    }

    // Set the bit at the given location
    public void BitSet(int location)
    {
        int n_byte;
        byte n_bit;
        n_byte = location / 8;

        n_bit = (byte) (Math.pow(2, (location % 8)));
        if (n_byte <= lastByte)
        {
            this.n[n_byte] = (byte) (this.n[n_byte] | n_bit);
        }
    }

    // Test to see if bit at the given location is set or not
    public boolean BitTest(int location)
    {
        int n_byte;
        byte n_bit;
        n_byte = location / 8;

        n_bit = (byte) (Math.pow(2, (location % 8)));

        if (n_byte <= lastByte)
        {
            if ((this.n[n_byte] & n_bit) != 0)
            {
                return true;
            }
        }
        return false;
    }

    /*
     * ************************** Private Methods ****************************
     */

    // Clone the BigInt value
    private BigInt Copy()
    {
        BigInt ans = new BigInt();

        // start with copying field variables
        ans.bitSize = this.bitSize;
        ans.lastByte = this.lastByte;
        ans.pfvArithMode = this.pfvArithMode;
        ans.pfvCarryBit = this.pfvCarryBit;
        ans.pfvLossOfPrecision = this.pfvLossOfPrecision;
        ans.pfvOverflow = this.pfvOverflow;

        // Now we make a copy of the array
        ans.n = new byte[lastByte + 1];
        for (int i = 0; i <= lastByte; i++)
        {
            ans.n[i] = n[i];
        }
        return ans;
    }

    // Set/Reset the bit size
    private void SetSize(int size)
    {
        if (size != bitSize)
        {
            BigInt temp = null;
            if ((size > bitSize) && (SignBit() && (pfvArithMode != ArithMode.Unsigned)))
            {
                temp = CreateMask((size - bitSize), true, size, pfvArithMode);
            }
            lastByte = (size - 1) / 8;
            this.n = ArrayResize(this.n, lastByte + 1);
            bitSize = size;
            Mask();

            // apply the extended sign mask
            if (temp != null)
            {
                for (int i = 0; i <= lastByte; i++)
                {
                    this.n[i] = (byte) (this.n[i] | temp.n[i]);
                }
            }
        }
    }

    // Add an extra byte for padding
    private void AddPadding(int size)
    {
        lastByte = (size - 1) / 8;

        this.n = ArrayResize(this.n, lastByte + 1);
        bitSize = size;
        Mask();
    }

    // Find the location of the most significant bit
    private int MaxBit()
    {
        for (int i = lastByte; i >= 0; i--)
        {
            for (int j = 7; j >= 0; j--)
            {
                if ((n[i] & (byte) Math.pow(2, j)) != 0)
                {
                    return ((i * 8) + j);
                }
            }
        }
        // Must be zero
        return 0;
    }

    // Reset the BigInt to Zero
    private void Clear()
    {
        for (int i = 0; i <= lastByte; i++)
        {
            n[i] = 0;
        }
        // Also reset the status flags
        pfvCarryBit = false;
        pfvOverflow = false;
        pfvLossOfPrecision = false;
    }

    // Returns the location of the most significant digit (byte) in the array
    private int MaxByte()
    {
        for (int i = lastByte; i >= 0; i--)
        {
            if (n[i] != 0)
            {
                return i;
            }
        }
        return 0;
    }

    // Put the integer value at the specified starting bit location within the
    // array
    private void PutAt(int i, int Location)
    {
        int n_bit, n_byte;
        long temp;

        n_byte = Location / 8;
        n_bit = Location % 8;

        temp = (long) (i * Math.pow(2, n_bit));

        n[n_byte] = (byte) (temp % 256);
        if (temp > 255)
        {
            if (n_byte < lastByte)
            {
                n[n_byte + 1] = (byte) (temp / 256);
            }
            else
            {
                pfvOverflow = true;
            }
        }
    }

    // Get an integer value from the specified location (and length) within the
    // array
    private int GetAt(int Location, int Length)
    {
        int ans, temp, n_bit, n_byte;

        n_byte = Location / 8;
        n_bit = (Location % 8);

        if (n_byte < lastByte)
        {
            temp = (int) (n[n_byte] & 0xff) + (int) (n[n_byte + 1] & 0xff) * 256;
        }
        else
        {
            temp = (int) (n[n_byte] & 0xff);
        }
        ans = temp / (int) Math.pow(2, n_bit) % (int) (Math.pow(2, Length));

        return ans;
    }

    // Mask the current value to the proper number of bits
    private void Mask()
    {
        byte mask_value;

        // Anything to do?
        if (bitSize == (lastByte + 1) * 8)
        {
            return;
        }
        mask_value = (byte) ((Math.pow(2, (bitSize - 1) % 8) * 2) - 1);

        n[lastByte] = (byte) (n[lastByte] & mask_value);
    }

    // Check to see if the Sign Bit is set
    private boolean SignBit()
    {
        byte mask_value;
        mask_value = (byte) (Math.pow(2, (bitSize - 1) % 8));

        if ((n[lastByte] & mask_value) != 0)
        {
            return true;
        }
        return false;
    }

    // Round up the bitsize to the next byte and add 1 byte
    private static int RoundUp(int bit_size)
    {
        return (((bit_size - 1) / 8) + 2) * 8;
    }

    // import a byte array into a BigInt
    private void Import(byte[] bytes, int size, boolean negative)
    {
        int last;

        SetSize(size);

        last = Math.min(lastByte, bytes.length - 1);

        for (int i = 0; i <= last; i++)
        {
            this.n[i] = bytes[i];
        }

        // Do we have any leftover bytes?
        pfvLossOfPrecision = false;
        if (last < bytes.length - 1)
        {
            for (int i = last + 1; i < bytes.length; i++)
            {
                if (bytes[i] != 0)
                {
                    pfvLossOfPrecision = true;
                    break;
                }
            }
        }

        // How about any leftover bits?
        if (!pfvLossOfPrecision)
        {
            pfvLossOfPrecision = (MaxBit() > bitSize);
        }

        // Fill in the missing sign extension
        if ((bytes.length - 1 < lastByte) && negative)
        {
            for (int i = bytes.length; i <= lastByte; i++)
            {
                this.n[i] = (byte) 255;
            }
        }
        Mask();
    }

    /*
     * These are a substitue for BitConverter in other languages
     */
    // convert a byte array into an integer
    private static int GetInt(byte[] b)
    {
        java.io.ByteArrayInputStream bais = new java.io.ByteArrayInputStream(reverse(b));
        java.io.DataInputStream stream = new java.io.DataInputStream(bais);

        try
        {
            return stream.readInt();
        }
        catch (Exception ignored) {}
        return 0;
    }

    // convert a byte array into a long
    private static long GetLong(byte[] b)
    {
        java.io.ByteArrayInputStream bais = new java.io.ByteArrayInputStream(reverse(b));
        java.io.DataInputStream stream = new java.io.DataInputStream(bais);

        try
        {
            return stream.readLong();
        }
        catch (Exception ignored) {}
        return 0;
    }

    // convert an integer into a byte array
    private static byte[] GetBytes(int value)
    {
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        java.io.DataOutputStream stream = new java.io.DataOutputStream(baos);

        try
        {
            stream.writeInt(value);
            stream.close();
        }
        catch (Exception ignored) {}
        return reverse(baos.toByteArray());
    }

    // convert a long into a byte array
    private static byte[] GetBytes(long value)
    {
        java.io.ByteArrayOutputStream baos = new java.io.ByteArrayOutputStream();
        java.io.DataOutputStream stream = new java.io.DataOutputStream(baos);
        try
        {
            stream.writeLong(value);
        }
        catch (Exception ignored) {}
        return reverse(baos.toByteArray());
    }

    // reverse the order of the bytes in the array
    private static byte[] reverse(byte[] arr)
    {
        byte[] rev = new byte[arr.length];

        for (int i = 0; i < arr.length; i++)
        {
            rev[arr.length - i - 1] = arr[i];
        }
        return rev;
    }

    private byte[] ArrayResize(byte[] b, int size)
    {
        byte[] ans = new byte[size];

        int len = Math.min(b.length, size);
        System.arraycopy(b, 0, ans, 0, len);
        return ans;
    }
}
