 package com.emmetgray.wrpn;

import android.util.Log;
import android.util.Xml;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlSerializer;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.Locale;
import java.util.Properties;

 /*
 * The state of the calculator.  These setting are what get stored in the internal
 * "state" during device rotations and intents. Plus this identical thing gets stored in
 * the "external" state between runs.
 */

public class CalcState
{
    private final int MIN_FLOAT = 0;
    private final int MAX_FLOAT = 9;
    private final int KeyDp = 0x48;
    private final int MIN_WORD = 1;
    private final int NUM_FLAGS = 6;

    // The major operating modes of the calculator
    public enum CalcOpMode
    {
        Float,
        Hex,
        Dec,
        Oct,
        Bin
    }

    // The calculator flags
    public enum CalcFlag
    {
        User0,
        User1,
        User2,
        LeadingZero,
        Carry,
        Overflow;

        // make a java enum a bit more civilized
        static private CalcFlag[] pList = null;

        public static CalcFlag fromVal(int val)
        {
            // a simple caching routine
            if (pList == null)
            {
                pList = CalcFlag.values();
            }

            if (val >= 0 && val < pList.length)
            {
                return pList[val];
            }
            // should this throw an exception?
            return User0;
        }
    }

    // In the real calculator the number of registers and program memory
    // "compete" for the same 203 bytes. So the max you can have are:
    // 1) 406 registers (at 4-bit word size) and 0 program memory, or
    // 2) 0 registers and 203 program memory
    //
    // However, in this version, there is no competition and the numbers
    // are static. You can choose any numbers you like in the config file.
    // The default is 32 registers and 203 lines of memory.

    public static final String TAG = "WRPN_log";
    public static Properties prop;

    private boolean pfvSaveOnExit;
    private int pfvWordSize;
    private CalcOpMode pfvOpMode;
    private BigInt.ArithMode pfvArithMode;
    private int pfvFloatPrecision;
    private boolean[] pfvFlags;
    private Register[] pfvReg;
    private Register pfvRegIndex;
    private CStack pfvStack;
    private Register pfvRegLastX;
    private int pfvPrgmPosition;
    private java.util.ArrayList<String> pfvPrgmMemory;
    private java.util.Stack<Integer> pfvPrgmRetStack;
    private boolean pfvPrgmRunning;

    private int numRegs;
    private boolean syncConv;
    private int numLines;

    // TODO: Add the ability to save state (during device rotation) in the PRGM mode

    // Default constructor
    public CalcState(String configFile)
    {
        // TODO: Only change the version when the properties file changes
        String propVersion = "7.0.2";

        /*
         * I use the Java "Properites" class to store seldom changed features.
         * This is not at all common in Android, but it *is* consistent with
         * the other versions of WRPN in different languages/platforms
         */
        prop = new Properties();

        // load the properties from the external file

        File config = new File(configFile);
        try
        {
            if (config.exists())
            {
                prop.loadFromXML(new FileInputStream(config));
            }
        }
        catch (Exception ex)
        {
            if (BuildConfig.DEBUG)
            {
                Log.e(TAG, "Properties: " + ex.toString());
            }
            // set the "Version" to a bogus number
            prop.setProperty("Version", "0");
        }

        Log.e("Prop", prop.getProperty("Version"));

        // doesn't exist or not the right version, then create/overwrite it
        if (!config.exists() || !prop.getProperty("Version").equals(propVersion))
        {
            prop.setProperty("NumRegisters", "32");
            prop.setProperty("PrgmMemoryLines", "302");
            prop.setProperty("SleepDelay", "1500");
            prop.setProperty("SyncConversions", "true");
            prop.setProperty("HomeURL", "https://www.wrpn.emmet-gray.com");
            prop.setProperty("Email", "egray1@hot.rr.com");
            prop.setProperty("Version", propVersion);
            prop.setProperty("HelpURL", "https://www.wrpn.emmet-gray.com/UsersGuide.html");
            prop.setProperty("StoreURL", "https://play.google.com/store/apps/details?id=com.emmetgray.wrpn"); // v7.0.2
            prop.setProperty("WebURL", "https://www.wrpn.emmet-gray.com/FAQ.html#Privacy"); // v7.0.2
            prop.setProperty("Orientation", "Auto");
            prop.setProperty("FontSize", "0"); // v7.0.0

            try
            {
                prop.storeToXML(new FileOutputStream(config), null);
            }
            catch (Exception ignored) {}
        }

        // cache these property values
        numRegs = Integer.parseInt(prop.getProperty("NumRegisters"));
        syncConv = Boolean.parseBoolean(prop.getProperty("SyncConversions"));
        numLines = Integer.parseInt(prop.getProperty("PrgmMemoryLines"));

        pfvFlags = new boolean[NUM_FLAGS];
        pfvReg = new Register[numRegs];

        // set the initial default conditions
        pfvSaveOnExit = true;
        pfvWordSize = 16;
        pfvOpMode = CalcOpMode.Float;
        pfvArithMode = BigInt.ArithMode.TwosComp;
        pfvFloatPrecision = 3;
        for (int i = 0; i < NUM_FLAGS; i++)
        {
            pfvFlags[i] = false;
        }
        pfvFlags[3] = true; // leading zeros
        for (int i = 0; i < numRegs; i++)
        {
            pfvReg[i] = new Register(pfvWordSize, pfvArithMode);
        }
        pfvRegIndex = new Register(64, pfvArithMode); // RegIndex has a fixed size
        pfvStack = new CStack(pfvWordSize, pfvArithMode);
        pfvRegLastX = new Register(pfvWordSize, pfvArithMode);
        pfvPrgmPosition = 0;
        pfvPrgmMemory = new java.util.ArrayList<String>();
        pfvPrgmRetStack = new java.util.Stack<Integer>();
        pfvPrgmRunning = false;
    }

    // Save the configuration on Exit
    public boolean getSaveOnExit()
    {
        return pfvSaveOnExit;
    }

    public void setSaveOnExit(boolean save)
    {
        pfvSaveOnExit = save;
    }

    // The current operating mode of the calculator
    public CalcOpMode getOpMode()
    {
        return pfvOpMode;
    }

    public void setOpMode(CalcOpMode opmode)
    {
        pfvOpMode = opmode;
    }

    // The Arithmetic mode (the complement mode)
    public BigInt.ArithMode getArithMode()
    {
        return pfvArithMode;
    }

    public void setArithMode(BigInt.ArithMode mode)
    {
        if (pfvArithMode != mode)
        {
            // change the Arithmetic Mode of everything
            ReArithAll(mode);
        }
        pfvArithMode = mode;
    }

    // The number of decimal points for the display
    public int getFloatPrecision()
    {
        return pfvFloatPrecision;
    }

    public void setFloatPrecision(int fp)
    {
        if (fp >= MIN_FLOAT && (fp <= MAX_FLOAT || fp == KeyDp))
        {
            pfvFloatPrecision = fp;
        }
    }

    // The size of the integers
    public int getWordSize()
    {
        return pfvWordSize;
    }

    public void setWordSize(int size)
    {
        if (size >= MIN_WORD)
        {
            if (pfvWordSize != size)
            {
                // change the size of everything
                ReSizeAll(size);
            }
            pfvWordSize = size;
        }
    }

    // The 6 user and systems flags
    public boolean getFlag(CalcFlag flag)
    {
        return pfvFlags[flag.ordinal()];
    }

    public void setFlag(CalcFlag flag, boolean val)
    {
        pfvFlags[flag.ordinal()] = val;
    }

    // The storage registers
    public Register getReg(int index)
    {
        // TODO: Any practical input validation here?
        return pfvReg[index];
    }

    public void setReg(int index, Register reg)
    {
        if (index >= 0 && index < numRegs)
        {
            pfvReg[index] = reg;
        }
    }

    // The index register
    public Register getRegIndex()
    {
        return pfvRegIndex;
    }

    public void setRegIndex(Register i)
    {
        pfvRegIndex = i;
    }

    // The Last X register
    public Register getRegLastX()
    {
        return pfvRegLastX;
    }

    public void setRegLastX(Register lx)
    {
        pfvRegLastX = lx;
    }

    // The location (the line) where the program is currently executing
    public int getPrgmPosition()
    {
        return pfvPrgmPosition;
    }

    public void setPrgmPosition(int pp)
    {
        if (pp >= 0 && pp <= numLines)
        {
           pfvPrgmPosition = pp;
        }
    }

    // Program memory
    public java.util.ArrayList<String> getPrgmMemory()
    {
        return pfvPrgmMemory;
    }

    // Program Return Stack
    public java.util.Stack<Integer> getPrgmRetStack()
    {
        return pfvPrgmRetStack;
    }

    // The Calculator Stack (x, y, z, and t registers)
    public CStack getStack()
    {
        return pfvStack;
    }

    // Is the calculator currently running a program
    public boolean getPrgmRunning()
    {
        return pfvPrgmRunning;
    }

    public void setPrgmRunning(boolean pr)
    {
        pfvPrgmRunning = pr;
    }

    // just a short cut for incrementing the program position
    public void incrementPrgrmPosition()
    {
        pfvPrgmPosition++;
    }

    /*
     * I'm using a custom XML serialization technique.  This isn't optimal
     * for Android, but it is compatible with the other versions of WRPN on
     * other platforms... and it produces pretty XML files!
     */

    // Save the CalcState as an XML string
    public String serialize() throws IllegalArgumentException, IllegalStateException, IOException
    {
        /*
         * This "low-level" XML parser technique is an attempt to make this perform similarly
         * to the more appropriate Parcel/Bundle technique
         */
        XmlSerializer s = Xml.newSerializer();
        s.setFeature("http://xmlpull.org/v1/doc/features.html#indent-output", true);
        StringWriter sw = new StringWriter();

        s.setOutput(sw);
        s.startDocument("UTF-8", true);

        // create the root element
        s.startTag("", "CalcState");
        s.attribute("", "saved", new java.util.Date().toString());

        s.comment("WRPN CalcState v" + prop.getProperty("Version").substring(0, 3));

        s.startTag("", "SaveOnExit");
        s.text(Boolean.toString(pfvSaveOnExit));
        s.endTag("", "SaveOnExit");

        s.startTag("", "WordSize");
        s.text(Integer.toString(pfvWordSize));
        s.endTag("", "WordSize");

        s.startTag("", "OpMode");
        s.text(pfvOpMode.toString());
        s.endTag("", "OpMode");

        s.startTag("", "ArithMode");
        s.text(pfvArithMode.toString());
        s.endTag("", "ArithMode");

        s.startTag("", "FloatPrecision");
        s.text(Integer.toString(pfvFloatPrecision));
        s.endTag("", "FloatPrecision");

        String[] FlagNames = {"User0", "User1", "User2", "LeadingZero", "Carry", "Overflow"};
        s.startTag("", "Flags");
        for (int i = 0; i < NUM_FLAGS; i++)
        {
            s.startTag("", "Flag");
            s.attribute("", "name", FlagNames[i]);
            s.text(Boolean.toString(pfvFlags[i]));
            s.endTag("", "Flag");
        }
        s.endTag("", "Flags");

        String[] RegNames = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9",
                "A", "B", "C", "D", "E", "F", ".0", ".1", ".2", ".3", ".4",
                ".5", ".6", ".7", ".8", ".9", ".A", ".B", ".C", ".D", ".E",
                ".F"};
        s.startTag("", "Regs");
        for (int i = 0; i < numRegs; i++)
        {
            s.startTag("", "Reg");
            if (i < RegNames.length)
            {
                s.attribute("", "name", RegNames[i]);
            }
            else
            {
                s.attribute("", "name", "Reg" + i);
            }
            s.startTag("", "FVal");
            s.text(pfvReg[i].getFVal().toString());
            s.endTag("", "FVal");
            s.startTag("", "BiVal");
            s.text(pfvReg[i].getBiVal().ToStringHex());
            s.endTag("", "BiVal");
            s.endTag("", "Reg");
        }
        s.endTag("", "Regs");

        s.startTag("", "RegIndex");
        s.startTag("", "FVal");
        s.text(pfvRegIndex.getFVal().toString());
        s.endTag("", "FVal");
        s.startTag("", "BiVal");
        s.text(pfvRegIndex.getBiVal().ToStringHex());
        s.endTag("", "BiVal");
        s.endTag("", "RegIndex");

        String[] StackName = {"T", "Z", "Y", "X"};
        Register[] StackArray = pfvStack.toArray();
        s.startTag("", "Stacks");
        for (int i = 0; i < 4; i++)
        {
            s.startTag("", "Stack");
            s.attribute("", "name", StackName[i]);
            s.startTag("", "FVal");
            s.text(StackArray[i].getFVal().toString());
            s.endTag("", "FVal");
            s.startTag("", "BiVal");
            s.text(StackArray[i].getBiVal().ToStringHex());
            s.endTag("", "BiVal");
            s.endTag("", "Stack");
        }
        s.endTag("", "Stacks");

        s.startTag("", "RegLastX");
        s.startTag("", "FVal");
        s.text(pfvRegLastX.getFVal().toString());
        s.endTag("", "FVal");
        s.startTag("", "BiVal");
        s.text(pfvRegLastX.getBiVal().ToStringHex());
        s.endTag("", "BiVal");
        s.endTag("", "RegLastX");

        s.startTag("", "PrgmPosition");
        s.text(Integer.toString(pfvPrgmPosition));
        s.endTag("", "PrgmPosition");

        // start with 1, since that's what the display uses
        int num = 1;
        s.startTag("", "PrgmMemory");
        for (String line : pfvPrgmMemory)
        {
            s.startTag("", "Line");
            s.attribute("", "name", String.format(Locale.US, "%03d", num));
            s.text(line);
            s.endTag("", "Line");
            num++;
        }
        s.endTag("", "PrgmMemory");

        num = 0;
        s.startTag("", "PrgmRetStack");
        if (pfvPrgmRetStack.size() > 0)
        {
            for (int i : pfvPrgmRetStack.toArray(new Integer[0]))
            {
                s.startTag("", "Return");
                s.attribute("", "name", Integer.toString(num));
                s.text(Integer.toString(i));
                s.endTag("", "Return");
                num++;
            }
        }
        s.endTag("", "PrgmRetStack");
        s.endTag("", "CalcState");
        s.endDocument();

        return sw.toString();
    }

    // Convert a saved XML file into the CalcState
    public void deserialize(String state) throws IOException, XmlPullParserException
    {
        /*
         * This was designed to be very simple (and efficient), but doing the parsing
         * "by hand" means that there is absolutely no sanity checking at all!
         */
        XmlPullParser xpp = Xml.newPullParser();
        xpp.setInput(new StringReader(state));

        int eventType = xpp.getEventType();
        boolean done = false;
        int index = 0;
        Register reg = new Register();
        String parentTag = "";

        while (eventType != XmlPullParser.END_DOCUMENT && !done)
        {
            String name = null;
            switch (eventType)
            {
                case XmlPullParser.START_DOCUMENT:
                    break;
                case XmlPullParser.START_TAG:
                    name = xpp.getName();
                    switch(name)
                    {
                        case "SaveOnExit":
                            pfvSaveOnExit = Boolean.parseBoolean(xpp.nextText());
                            break;
                        case "WordSize":
                            pfvWordSize = Integer.parseInt(xpp.nextText());
                            break;
                        case "OpMode":
                            pfvOpMode = CalcOpMode.valueOf(xpp.nextText());
                            break;
                        case "ArithMode":
                            pfvArithMode = BigInt.ArithMode.valueOf(xpp.nextText());
                            break;
                        case "FloatPrecision":
                            pfvFloatPrecision = Integer.parseInt(xpp.nextText());
                            break;
                        case "Flags":
                            index = 0;
                            break;
                        case "Flag":
                            pfvFlags[index] = Boolean.parseBoolean(xpp.nextText());
                            // increment the flag count
                            index++;
                            break;
                        case "Regs":
                            parentTag = "Regs";
                            index = 0;
                            break;
                        case "FVal":
                            double f = Double.parseDouble(xpp.nextText());
                            switch(parentTag)
                            {
                                case "Regs":
                                    pfvReg[index].setFVal(f);
                                    break;
                                case "RegIndex":
                                    pfvRegIndex.setFVal(f);
                                    break;
                                case "Stack":
                                    reg.setFVal(f);
                                    break;
                                case "RegLastX":
                                    pfvRegLastX.setFVal(f);
                                    break;
                            }
                            break;
                        case "BiVal":
                            switch (parentTag)
                            {
                                case "Regs":
                                    pfvReg[index].setBiVal(new BigInt("&H" + xpp.nextText(), pfvWordSize, pfvArithMode));
                                    // increment the reg count
                                    index++;
                                    break;
                                case "RegIndex":
                                    // word size is always 64
                                    pfvRegIndex.setBiVal(new BigInt("&H" + xpp.nextText(), 64, pfvArithMode));
                                    break;
                                case "Stack":
                                    reg.setBiVal(new BigInt("&H" + xpp.nextText(), pfvWordSize, pfvArithMode));
                                    pfvStack.push(reg);
                                    break;
                                case "RegLastX":
                                    pfvRegLastX.setBiVal(new BigInt("&H" + xpp.nextText(), pfvWordSize, pfvArithMode));
                                    break;
                            }
                            break;
                        case "RegIndex":
                            parentTag = "RegIndex";
                            index = 0;
                            break;
                        case "Stacks":
                            parentTag = "Stack";
                            break;
                        case "RegLastX":
                            parentTag = "RegLastX";
                            break;
                        case "PrgmPosition":
                            pfvPrgmPosition = Integer.parseInt(xpp.nextText());
                            break;
                        case "PrgmMemory":
                            pfvPrgmMemory.clear();
                            break;
                        case "Line":
                            pfvPrgmMemory.add(xpp.nextText());
                            break;
                        case "PrgmRetStack":
                            pfvPrgmRetStack.clear();
                            break;
                        case "Return":
                            index = Integer.parseInt(xpp.nextText());
                            pfvPrgmRetStack.push(index);
                            break;
                    }
                    break;
                case XmlPullParser.END_TAG:
                    name = xpp.getName();
                    if (name.equals("CalcState"))
                    {
                        done = true;
                    }
                    break;
            }
            eventType = xpp.next();
        }
    }

    // Resize all of the Big Integer values inside the registers
    private void ReSizeAll(int size)
    {
        if (syncConv)
        {
            for (int i = 0; i < numRegs; i++)
            {
                pfvReg[i].getBiVal().setWordSize(size);
            }
            pfvRegLastX.getBiVal().setWordSize(size);
            pfvStack.getX().getBiVal().setWordSize(size);
            pfvStack.getY().getBiVal().setWordSize(size);
            pfvStack.getZ().getBiVal().setWordSize(size);
            pfvStack.getT().getBiVal().setWordSize(size);
        }
        else
        {
            // Using the rules on page 32 and 66 the storage registers
            // are not resized. Also, the stack does not preserve the
            // sign bit on widening
            BigInt.ArithMode tempArith = pfvArithMode;
            pfvRegLastX.getBiVal().setBIArithMode(BigInt.ArithMode.Unsigned);
            pfvStack.getX().getBiVal().setBIArithMode(BigInt.ArithMode.Unsigned);
            pfvStack.getY().getBiVal().setBIArithMode(BigInt.ArithMode.Unsigned);
            pfvStack.getZ().getBiVal().setBIArithMode(BigInt.ArithMode.Unsigned);
            pfvStack.getT().getBiVal().setBIArithMode(BigInt.ArithMode.Unsigned);

            pfvRegLastX.getBiVal().setWordSize(size);
            pfvStack.getX().getBiVal().setWordSize(size);
            pfvStack.getY().getBiVal().setWordSize(size);
            pfvStack.getZ().getBiVal().setWordSize(size);
            pfvStack.getT().getBiVal().setWordSize(size);

            pfvRegLastX.getBiVal().setBIArithMode(tempArith);
            pfvStack.getX().getBiVal().setBIArithMode(tempArith);
            pfvStack.getY().getBiVal().setBIArithMode(tempArith);
            pfvStack.getZ().getBiVal().setBIArithMode(tempArith);
            pfvStack.getT().getBiVal().setBIArithMode(tempArith);
        }
    }

    // Reset the Arithmetic mode of all of the BigInt values inside the
    // registers
    private void ReArithAll(BigInt.ArithMode mode)
    {
        for (int i = 0; i < numRegs; i++)
        {
            pfvReg[i].getBiVal().setBIArithMode(mode);
        }
        pfvRegLastX.getBiVal().setBIArithMode(mode);
        pfvRegIndex.getBiVal().setBIArithMode(mode);
        pfvStack.getX().getBiVal().setBIArithMode(mode);
        pfvStack.getY().getBiVal().setBIArithMode(mode);
        pfvStack.getZ().getBiVal().setBIArithMode(mode);
        pfvStack.getT().getBiVal().setBIArithMode(mode);
    }

    // Synchronize the Big Integer and Float values within a Register
    public void SyncValues()
    {
        if (syncConv)
        {
            // With SyncConversion set to true, we do NOT follow the behavior
            // of the real calculator. Instead, we synchronize the integer
            // and float values when switching between modes. Obviously, there
            // will likely be some loss of precision when decimal values are
            // truncated.
            if (getOpMode() == CalcOpMode.Float)
            {
                // if we're currently in the float mode, then we need to
                // copy those values to the to the integer mode
                for (int i = 0; i < numRegs; i++)
                {
                    pfvReg[i].setBiVal(new BigInt(pfvReg[i].getFVal(), pfvWordSize, pfvArithMode));
                }
                pfvRegLastX.setBiVal(new BigInt(pfvRegLastX.getFVal(), pfvWordSize, pfvArithMode));
                pfvStack.getX().setBiVal(new BigInt(pfvStack.getX().getFVal(), pfvWordSize, pfvArithMode));
                pfvStack.getY().setBiVal(new BigInt(pfvStack.getY().getFVal(), pfvWordSize, pfvArithMode));
                pfvStack.getZ().setBiVal(new BigInt(pfvStack.getZ().getFVal(), pfvWordSize, pfvArithMode));
                pfvStack.getT().setBiVal(new BigInt(pfvStack.getT().getFVal(), pfvWordSize, pfvArithMode));
                pfvRegIndex.setBiVal(new BigInt(pfvRegIndex.getFVal(), 64, pfvArithMode));
            }
            else
            {
                // copy the values to the float mode
                for (int i = 0; i < numRegs; i++)
                {
                    pfvReg[i].setFVal(pfvReg[i].getBiVal().ToLong());
                }
                pfvRegLastX.setFVal(pfvRegLastX.getBiVal().ToLong());
                pfvStack.getX().setFVal(pfvStack.getX().getBiVal().ToLong());
                pfvStack.getY().setFVal(pfvStack.getY().getBiVal().ToLong());
                pfvStack.getZ().setFVal(pfvStack.getZ().getBiVal().ToLong());
                pfvStack.getT().setFVal(pfvStack.getT().getBiVal().ToLong());
                pfvRegIndex.setFVal(pfvRegIndex.getBiVal().ToLong());
            }
        }
        else
        {
            // If no "Sync Conversion" then we use the rules from the real
            // calculator (yuk!)
            if (getOpMode() == CalcOpMode.Float)
            {
                // The rules for converting from Float mode to Integer mode
                // are found at page 60
                pfvWordSize = 56;

                double x, temp;
                long tempX, tempY;
                boolean neg;

                tempX = 0;
                tempY = 0;
                neg = false;
                if (pfvStack.getX().getFVal() != 0.0)
                {
                    x = pfvStack.getX().getFVal();
                    if (x < 0)
                    {
                        neg = true;
                        x = x * -1;
                    }
                    temp = x / Math.pow(2, 32);
                    tempX = (long) (Math.log(temp) / Math.log(2));
                    tempY = (long) (x / Math.pow(2, tempX));
                    if (neg)
                    {
                        tempY = tempY * -1;
                    }
                }

                pfvStack.getX().setBiVal(new BigInt(tempX, pfvWordSize, pfvArithMode));
                pfvStack.getY().setBiVal(new BigInt(tempY, pfvWordSize, pfvArithMode));
            }
            else
            {
                // The rules for converting from Integer mode to Float mode
                // are found at page 57-58
                pfvStack.getY().setFVal(0);
                pfvStack.getZ().setFVal(0);
                pfvStack.getT().setFVal(0);
                pfvRegLastX.setFVal(0);

                // set x = (y)(2^x)
                pfvStack.getX().setFVal(pfvStack.getY().getBiVal().ToLong() * Math.pow(2, pfvStack.getX().getBiVal().ToLong()));
            }
        }
    }
}